--[[

Skillet: A tradeskill window replacement.
Copyright (c) 2007 Robert Clark <nogudnik@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

]]--

local AceEvent = AceLibrary("AceEvent-2.0")

local QUEUE_DEBUG = false

-- Adds the recipe to the queue of recipes to be processed. If the recipe
-- is already in the queue, then the count of items to be created is increased,
-- otherwise the recipe is added it the end
--
-- If there are item needs to make the recipe that are not currently in your
-- inventory, but you can craft them, then they are added to the queue before the
-- requested recipe.
local function add_items_to_queue(skillIndex, recipe, count)
    assert(tonumber(skillIndex) and recipe and tonumber(count),"Usage: add_items_to_queue(skillIndex, recipe, count)")

    -- if we need mats that are not in the inventory, but are craftable, add
    -- the mats to the queue first

    if QUEUE_DEBUG then
        Skillet:Print("Adding " .. count .. "x" .. recipe.link)
    end

    if Skillet.db.profile.queue_craftable_reagents then
        -- never more that 8 reagents
        for i=1, 8, 1 do
            reagent = recipe[i]

            if not reagent then
                break
            end

            local needed = (reagent.needed * count)
            local   have = GetItemCount(reagent.link, true)

            if QUEUE_DEBUG then
                Skillet:Print("  have " .. have .. "x" .. reagent.link .. ", need " .. needed)
            end

            if have < needed then
                local item = Skillet.stitch:GetItemDataByName(reagent.name)

                if item and item.link == reagent.link then
                    -- we can craft this
                    -- the extra check for an exact name match is because the
                    -- Stitch search will fall back on a wild card across all
                    -- skills if an exact match is not found
                    add_items_to_queue(item.index, item, (needed - have))
                end
            end
        end
    end

	Skillet.stitch:AddToQueue(skillIndex, count)

    -- XXX: This is a bit hacky, try to think of something smarter
    Skillet:SaveQueue(Skillet.db.server.queues, Skillet.currentTrade)
end

-- Save the current queue into the provided database
function Skillet:SaveQueue(db, tradeskill)
    if not db[UnitName("player")] then
        db[UnitName("player")] = {}
    end

    db[UnitName("player")][tradeskill] = self.stitch.queue
end

-- Loads the queue for the provided tradeskill name from the database
function Skillet:LoadQueue(db, tradeskill)
    if not db[UnitName("player")] then
        db[UnitName("player")] = {}
    end
    if not db[UnitName("player")][tradeskill] then
        db[UnitName("player")][tradeskill] = {}
    end

    self.stitch.queue = db[UnitName("player")][tradeskill]

    AceEvent:TriggerEvent("SkilletStitch_Queue_Add")
end

-- Queue the max number of craftable items for the currently selected skill
function Skillet:QueueAllItems()
	if self.currentTrade and self.selectedSkill then
		local s = self.stitch:GetItemDataByIndex(self.currentTrade, self.selectedSkill)
        if s then
            local count = s.numcraftable - self.stitch:GetNumQueuedItems(self.selectedSkill)
            if count > 0 then
                add_items_to_queue(self.selectedSkill, s, count)
            end
			-- queued all that could be created, reset the create count
			-- back down to 0
			self:UpdateNumItemsSlider(0, false);
		end
	end
end

-- Adds the currently selected number of items to the queue
function Skillet:QueueItems()
	self.numItemsToCraft = SkilletItemCountInputBox:GetNumber();

	if self.numItemsToCraft > 0 then
		if self.currentTrade and self.selectedSkill then
			local s = self.stitch:GetItemDataByIndex(self.currentTrade, self.selectedSkill);
			if s then
				add_items_to_queue(self.selectedSkill, s, self.numItemsToCraft)
			end
		end
	end
end

-- Queue and create the max number of craftable items for the currently selected skill
function Skillet:CreateAllItems()
	if self.currentTrade and self.selectedSkill then
		local s = self.stitch:GetItemDataByIndex(self.currentTrade, self.selectedSkill);
        if s then
            local count = s.numcraftable - self.stitch:GetNumQueuedItems(self.selectedSkill)
            if count > 0 then
                add_items_to_queue(self.selectedSkill, s, count)
                self:ProcessQueue()
            end
            -- created all that could be created, reset the create count
            -- back down to 0
            self:UpdateNumItemsSlider(0, false)
		end
	end
end

-- Adds the currently selected number of items to the queue and then starts the queue
function Skillet:CreateItems()
	self.numItemsToCraft = SkilletItemCountInputBox:GetNumber();

	if self.numItemsToCraft > 0 then
		if self.currentTrade and self.selectedSkill then
			local s = self.stitch:GetItemDataByIndex(self.currentTrade, self.selectedSkill);
			if s then
				add_items_to_queue(self.selectedSkill, s, self.numItemsToCraft)
				self:ProcessQueue();
			end
		end
	end
end

-- Starts the enchant process for the selected recipe item
-- This does not go through the Stitch library because it does
-- not support queuing enchants (mainly because queueing enchants is silly) .
function Skillet:EnchantItem()
	if self:IsCraft() and self.selectedSkill then
		DoCraft(self.selectedSkill)
	end
end

-- Starts Processing any items in the queue
function Skillet:ProcessQueue()
	local queue = self.stitch:GetQueueInfo()
	if not queue then
		return
	end

	self.stitch:ProcessQueue()
end

-- Clears the current queue, this will not cancel an
-- items currently being crafted.
function Skillet:EmptyQueue()
	self.stitch:ClearQueue()
end

-- Removes an item from the queue
function Skillet:RemoveQueuedItem(id)
    local queue = self.stitch:GetQueueInfo();
    if not queue then
        -- this should never happen, log an error?
        return
    end

    if id == 1 and not self:IsCraft() then
        self.stitch:CancelCast()
    end

    self.stitch:RemoveFromQueue(id)
    self:SaveQueue(self.db.server.queues, self.currentTrade)

    self:UpdateQueueWindow()
end

-- Returns a table {playername, queues} containing all queued
-- items
function Skillet:GetAllQueues()
    if not self.db.server.queues then
        return {}
    end

    return self.db.server.queues
end

-- Returns the list of queues for the specified player
function Skillet:GetQueues(player)
    assert(tostring(player),"Usage: GetQueues('player_name')")

    if not self.db.server.queues then
        return {}
    end

    if not self.db.server.queues[player] then
        return {}
    end

    return self.db.server.queues[player]
end

-- Returns the list of queues for the current player
function Skillet:GetPlayerQueues()
    return self:GetQueues(UnitName("player"))
end